# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from __future__ import annotations

import dataclasses as dtcl
from typing import Sequence, Tuple, Union

import numpy as nmpy
import vedo.colors as clrs
from mayavi import mlab

import cell_tracking_BC.in_out.graphics.dbe.mayavi.style as styl
import cell_tracking_BC.in_out.graphics.generic.any_d as gnrc
from cell_tracking_BC.in_out.graphics.context import axes_3d_t as abstract_axes_3d_t
from cell_tracking_BC.in_out.graphics.context import colormap_h
from cell_tracking_BC.in_out.graphics.dbe.mayavi.any_d import figure_t as base_figure_t


array_t = nmpy.ndarray


@dtcl.dataclass(init=False, repr=False, eq=False)
class figure_t(base_figure_t, abstract_axes_3d_t):
    @classmethod
    def NewFigureAndAxes(
        cls, /, *, n_rows: int = 1, n_cols: int = 1, title: str = None
    ) -> Tuple[figure_t, Union[axes_t, Sequence[axes_t], Sequence[Sequence[axes_t]]],]:
        """"""
        figure = cls(figure=title, bgcolor=(1, 1, 1))
        axes = figure

        mlab.axes(
            xlabel="row positions",
            ylabel="column positions",
            zlabel="time points",
            figure=figure,
        )
        mlab.outline(figure=figure)

        return figure, axes

    @staticmethod
    def TimeScaling(shape: Sequence[int], length: int, /) -> float:
        """"""
        size = 0.5 * sum(shape)

        return size / length

    @staticmethod
    def MillefeuilleScaling(shape: Sequence[int], length: int, /) -> float:
        """"""
        size = 0.5 * sum(shape)

        return size / length

    def AddColormapFromMilestones(
        self,
        name: str,
        milestones: Sequence[Tuple[float, str]],
        /,
        *,
        position: str = "right",
    ) -> colormap_h:
        """"""
        output = lambda _vle: gnrc.ZeroOneValueToRGBAWithMilestones(
            _vle, milestones, clrs.getColor
        )

        return output

    def PlotLines(self, xs, ys, *args, zdir="z", **kwargs):
        """"""
        zs = args[0]

        if "color" in kwargs:
            color, _ = gnrc.ColorAndAlpha(kwargs["color"], clrs.getColor)
        else:
            color = "black"

        # tube_radius=0.025*size
        mlab.plot3d(xs, ys, zs, color=color, figure=self)

    def PlotText(self, x, y, z, s, zdir=None, **kwargs):
        """"""
        mlab.text3d(
            x,
            y,
            z,
            s,
            figure=self,
            **styl.ConvertedTextStyle(kwargs),
        )


axes_t = figure_t



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from __future__ import annotations

import dataclasses as dtcl
from typing import Sequence, Tuple, Union

import numpy as nmpy
import vedo.colors as clrs
from vedo import Axes as vedo_axes_t
from vedo import Box as box_t
from vedo import Picture as picture_t
from vedo import Volume as volume_t
from vedo.shapes import Lines as lines_t
from vedo.shapes import Text3D as text_t

import cell_tracking_BC.in_out.graphics.dbe.vedo.style as styl
import cell_tracking_BC.in_out.graphics.generic.any_d as gnrc
from cell_tracking_BC.in_out.graphics.context import annotation_h
from cell_tracking_BC.in_out.graphics.context import axes_3d_t as base_axes_3d_t
from cell_tracking_BC.in_out.graphics.context import cell_annotation_h, colormap_h
from cell_tracking_BC.in_out.graphics.dbe.vedo.any_d import figure_t as figure_anyd_t


array_t = nmpy.ndarray


_MAX_N_TIME_LABELS = 15


@dtcl.dataclass(init=False, repr=False, eq=False)
class figure_t(figure_anyd_t, base_axes_3d_t):
    def SetTimeAxisProperties(self, latest: int, /) -> None:
        """"""
        origin, corner = self.bbox_origin_and_corner

        length, width, height = (_max - _min for _min, _max in zip(origin, corner))

        x_range, y_range, z_range = ((_min, _max) for _min, _max in zip(origin, corner))
        n_time_labels = min(latest + 1, _MAX_N_TIME_LABELS)
        time_labels = tuple(
            (_pos, _lbl)
            for _pos, _lbl in zip(
                nmpy.linspace(self.bbox[0][2], self.bbox[1][2], num=n_time_labels),
                nmpy.around(nmpy.linspace(0, latest, num=n_time_labels), decimals=1),
            )
        )

        box = box_t(pos=origin, length=length, width=width, height=height).alpha(0.1)
        axes = vedo_axes_t(
            box,
            xrange=x_range,
            yrange=y_range,
            zrange=z_range,
            zValuesAndLabels=time_labels,
            c="black",
        )

        self.__iadd__(axes)

    @staticmethod
    def TimeScaling(shape: Sequence[int], length: int, /) -> float:
        """"""
        size = 0.5 * sum(shape)

        return 2.0 * size / (length - 1)

    @staticmethod
    def MillefeuilleScaling(shape: Sequence[int], length: int, /) -> float:
        """"""
        size = 0.5 * sum(shape)

        return 0.1 * size

    def AddColormapFromMilestones(
        self,
        name: str,
        milestones: Sequence[Tuple[float, str]],
        /,
        *,
        position: str = "right",
    ) -> colormap_h:
        """"""
        # Leave it in 3-D (as opposed to any-D) because of 3-D box below
        output = lambda _vle: gnrc.ZeroOneValueToRGBAWithMilestones(
            _vle, milestones, clrs.getColor
        )

        discretized = tuple(
            output(_vle)[:3] for _vle in nmpy.linspace(0.0, 1.0, num=100)
        )
        box = box_t(pos=(0, 0, 0), length=10, width=10, height=10).alpha(0.0)
        box.cmap(discretized, box.points()[:, 2])
        box.addScalarBar(title=name, nlabels=0)

        self.__iadd__(box)

        return output

    def PlotLines(self, xs, ys, *args, zdir="z", **kwargs):
        """"""
        zs = args[0]

        if "color" in kwargs:
            color, alpha = gnrc.ColorAndAlpha(kwargs["color"], clrs.getColor)
        else:
            color, alpha = "black", None

        points = list(zip(ys, xs, zs))
        lines = lines_t(points[:-1], points[1:], c=color, lw=2)
        if alpha is not None:
            lines.alpha(alpha)

        self.__iadd__(lines)
        self.UpdateBBoxFromMany(ys, xs, zs)

    def PlotText(self, x, y, z, s, zdir=None, **kwargs):
        """"""
        position = [y, x, z]
        additionals = styl.ConvertedTextStyle(kwargs)
        if "alpha" in additionals:
            alpha = additionals["alpha"]
            del additionals["alpha"]
        else:
            alpha = None

        text = text_t(s, position, **additionals)
        if alpha is not None:
            text.alpha(alpha)

        text.followCamera()

        self.__iadd__(text)
        self.UpdateBBoxFromOne(position)

    def PlotCellAnnotation(
        self, position: Tuple[float, float, float], text: cell_annotation_h, /, **kwargs
    ) -> Union[annotation_h, Sequence[annotation_h]]:
        """"""
        additionals = kwargs.copy()
        if "alpha" in additionals:
            alpha = additionals["alpha"]
            del additionals["alpha"]
        else:
            alpha = None

        # TODO: multi-piece text is currently dealt with as single-piece text
        if not isinstance(text, str):
            text = "".join(_elm[0] for _elm in text)

        flipped = [position[1], position[0], position[2]]
        text = text_t(text, flipped, **additionals)
        if alpha is not None:
            text.alpha(alpha)
        text.followCamera()

        self.__iadd__(text)
        self.UpdateBBoxFromOne(flipped)

        return None

    def PlotImageInZ(
        self,
        image: array_t,
        all_rows: array_t,
        all_cols: array_t,
        n_levels: int,
        height: float,
        z_scaling: float,
        /,
        min_intensity: float = 0.0,
        intensity_range: float = 1.0,
        alpha: float = 0.8,
        **kwargs,
    ) -> None:
        """"""
        scaled = 255.0 * (image - min_intensity) / intensity_range
        picture = picture_t(scaled).z(z_scaling * height).alpha(alpha)

        self.__iadd__(picture)
        self.UpdateBBoxFromMany(all_cols.flatten(), all_rows.flatten(), (height,))

    def PlotIsosurface(
        self,
        volume: array_t,
        iso_value: float,
        /,
        should_be_capped: bool = False,
        keep_every: int = 2,
        **kwargs,
    ) -> None:
        """"""
        if should_be_capped:
            cap = nmpy.zeros_like(volume[0, ...])[None, ...]
            volume = nmpy.concatenate((cap, volume, cap), axis=0)

        isosurface = volume_t(volume).isosurface(threshold=[iso_value])
        time_scaling = self.__class__.TimeScaling(volume.shape[:2], volume.shape[2])
        isosurface.scale(s=(1.0, 1.0, time_scaling))

        self.__iadd__(isosurface)
        self.UpdateBBoxFromMany(
            isosurface.xbounds(), isosurface.ybounds(), isosurface.zbounds()
        )


axes_t = figure_t



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from __future__ import annotations

import dataclasses as dtcl
from pathlib import Path as path_t
from typing import Any, Callable, Dict
from typing import NamedTuple as named_tuple_t
from typing import Protocol, Sequence, Tuple, Type, TypeVar, Union

from numpy import ndarray as array_t

from cell_tracking_BC.in_out.file.archiver import archiver_t


annotation_h = TypeVar("annotation_h")
element_h = TypeVar("element_h")
event_h = TypeVar("event_h")
frame_h = TypeVar("frame_h")
key_event_h = TypeVar("key_event_h")
path_collection_h = TypeVar("path_collection_h")
s_viewer_2d_h = TypeVar("s_viewer_2d_h")
scroll_event_h = TypeVar("scroll_event_h")
slider_h = TypeVar("slider_h")
t_viewer_2d_h = TypeVar("t_viewer_2d_h")

# "some_text" or (("text1", properties1), ("text2", properties2)...)
cell_annotation_h = Union[str, Sequence[Tuple[str, Dict[str, Any]]]]
rgb_color_h = Tuple[float, float, float]
rgba_color_h = Tuple[float, float, float, float]
colormap_h = Callable[
    [Union[float, Sequence[float]]], Union[rgba_color_h, Sequence[rgba_color_h]]
]


@dtcl.dataclass(init=False, repr=False, eq=False)
class axes_t:
    def SetTitle(self, title: str, /) -> None:
        pass

    def SetYAxisTicks(
        self, positions: Sequence[float], labels: Sequence[str], /
    ) -> None:
        pass

    def SetTimeAxisProperties(self, latest: int, /) -> None:
        pass

    def TurnTicksOff(self) -> None:
        pass

    def Freeze(self) -> None:
        pass

    def ShowLegend(self) -> None:
        pass

    # def RemoveArtists(self, which: Union[type, Sequence[type]], /) -> None:
    #     pass

    def AddStandardColormap(
        self, name: str, colormap: str, /, *, position: str = "right"
    ) -> colormap_h:
        pass

    def AddColormapFromMilestones(
        self,
        name: str,
        milestones: Sequence[Tuple[float, str]],
        /,
        *,
        position: str = "right",
    ) -> colormap_h:
        pass

    def RemoveLinesAndAnnotations(self) -> None:
        pass

    def Figure(self) -> figure_t:
        pass


@dtcl.dataclass(init=False, repr=False, eq=False)
class axes_2d_t(axes_t):
    def SetAxesPropertiesForTracking(self, tick_positions: Sequence[float], /) -> None:
        pass

    def PlotPoints(
        self,
        x,
        y,
        s=None,
        c=None,
        marker=None,
        cmap=None,
        norm=None,
        vmin=None,
        vmax=None,
        alpha=None,
        linewidths=None,
        *,
        edgecolors=None,
        plotnonfinite=False,
        data=None,
        **kwargs,
    ):
        pass

    def PlotLines(self, *args, scalex=True, scaley=True, data=None, **kwargs):
        pass

    def PlotText(self, x, y, s, fontdict=None, **kwargs):
        pass

    def PlotAnnotation(self, text, xy, *args, **kwargs):
        pass

    def PlotCellAnnotation(
        self, position: Tuple[float, float], text: cell_annotation_h, /, **kwargs
    ) -> Union[annotation_h, Sequence[annotation_h]]:
        pass

    def PlotImage(
        self, image: array_t, /, *, interval: Tuple[float, float] = None
    ) -> None:
        """"""
        pass

    def UpdateImage(
        self,
        picture: array_t,
        /,
        *,
        interval: Tuple[float, float] = None,
        should_update_limits: bool = False,
    ) -> None:
        """"""
        pass


@dtcl.dataclass(init=False, repr=False, eq=False)
class axes_3d_t(axes_t):
    @staticmethod
    def TimeScaling(shape: Sequence[int], length: int, /) -> float:
        """"""
        return 1.0

    @staticmethod
    def MillefeuilleScaling(shape: Sequence[int], length: int, /) -> float:
        """"""
        return 1.0

    def PlotPoints(
        self, xs, ys, zs=0, zdir="z", s=20, c=None, depthshade=True, *args, **kwargs
    ):
        pass

    def PlotLines(self, xs, ys, *args, zdir="z", **kwargs):
        pass

    def PlotText(self, x, y, z, s, zdir=None, **kwargs):
        pass

    def PlotAnnotation(self, text, xyz, *args, **kwargs):
        pass

    def PlotCellAnnotation(
        self, position: Tuple[float, float, float], text: cell_annotation_h, /, **kwargs
    ) -> Union[annotation_h, Sequence[annotation_h]]:
        pass

    def PlotImageInZ(
        self,
        image: array_t,
        all_rows: array_t,
        all_cols: array_t,
        n_levels: int,
        height: float,
        z_scaling: float,
        /,
        min_intensity: float = 0.0,
        intensity_range: float = 1.0,
        alpha: float = 0.8,
        **kwargs,
    ) -> None:
        pass

    def PlotIsosurface(
        self,
        volume: array_t,
        iso_value: float,
        /,
        should_be_capped: bool = False,
        keep_every: int = 2,
        **kwargs,
    ) -> None:
        pass


@dtcl.dataclass(init=False, repr=False, eq=False)
class figure_t:
    @classmethod
    def NewFigureAndAxes(
        cls, /, *, n_rows: int = 1, n_cols: int = 1, title: str = None
    ) -> Tuple[figure_t, Union[axes_t, Sequence[axes_t], Sequence[Sequence[axes_t]]]]:
        pass

    def ActivateTightLayout(self, *, pad=1.08, h_pad=None, w_pad=None, rect=None):
        pass

    def Show(
        self,
        /,
        *,
        interactively: bool = True,
        in_main_thread: bool = True,
    ) -> None:
        pass

    @staticmethod
    def ShowAll(
        *,
        interactively: bool = True,
        in_main_thread: bool = True,
    ) -> None:
        pass

    def Update(self, /, *, gently: bool = True) -> None:
        pass

    def Content(self, /) -> array_t:
        pass

    def Save(self, path: Union[str, path_t], /) -> None:
        pass

    def Archive(
        self,
        /,
        *,
        name: str = "figure",
        archiver: archiver_t = None,
    ) -> None:
        pass

    def ActivateEvent(self, event: str, processor: Callable[[event_h], Any], /) -> None:
        pass

    def Close(self) -> None:
        pass


class _sg(Protocol):  # sg=signature
    @staticmethod
    def cell_annotation_style_h(
        highlighted: bool, multi_track: bool, /
    ) -> Dict[str, Any]:
        ...

    @staticmethod
    def new_slider_h(figure: figure_t, n_steps: int, /) -> slider_h:
        ...

    @staticmethod
    def slider_value_h(slider: slider_h, /) -> float:
        ...

    @staticmethod
    def update_slider_h(slider: slider_h, value: float, /) -> None:
        ...

    @staticmethod
    def slider_bounds_h(slider: slider_h, /) -> Sequence[float]:
        ...

    @staticmethod
    def slider_axes_h(slider: slider_h, /) -> axes_2d_t:
        ...

    @staticmethod
    def get_visibility_h(what: element_h, /) -> bool:
        ...

    @staticmethod
    def set_visibility_h(what: element_h, visibility: bool, /) -> None:
        ...

    @staticmethod
    def is_target_of_event_h(axes: axes_t, event: event_h, /) -> bool:
        ...

    @staticmethod
    def key_event_key_h(event: key_event_h, /) -> str:
        ...

    @staticmethod
    def scroll_event_step_h(event: scroll_event_h, /) -> float:
        ...


class context_t(named_tuple_t):
    figure_2d_t: Type[figure_t] = None
    figure_3d_t: Type[figure_t] = None
    axes_2d_t: Type[axes_2d_t] = None
    axes_3d_t: Type[axes_3d_t] = None
    s_viewer_2d_t: s_viewer_2d_h = None
    t_viewer_2d_t: t_viewer_2d_h = None
    CellAnnotationStyle: _sg.cell_annotation_style_h = lambda: {}
    NewSlider: _sg.new_slider_h = None
    SliderValue: _sg.slider_value_h = None
    UpdateSlider: _sg.update_slider_h = None
    SliderBounds: _sg.slider_bounds_h = None
    SliderAxes: _sg.slider_axes_h = lambda: None
    IsVisible: _sg.get_visibility_h = None
    SetVisibility: _sg.set_visibility_h = None
    IsTargetOfEvent: _sg.is_target_of_event_h = lambda: True
    KeyEventKey: _sg.key_event_key_h = None
    ScrollEventStep: _sg.scroll_event_step_h = None



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from typing import Optional, Union

import cell_tracking_BC.in_out.graphics.generic.any_d as gnrc
import cell_tracking_BC.in_out.graphics.generic.three_d as gnr3
from cell_tracking_BC.in_out.graphics.context import axes_3d_t, context_t, figure_t
from cell_tracking_BC.in_out.graphics.dbe.matplotlib.context import DRAWING_CONTEXT
from cell_tracking_BC.in_out.graphics.generic.two_d_sequence import s_viewer_2d_t
from cell_tracking_BC.type.segmentation import compartment_t
from cell_tracking_BC.type.segmentations import segmentations_t
from cell_tracking_BC.type.sequence import sequence_h, sequence_t


def ShowSegmentation(
    sequence: sequence_h,
    /,
    *,
    compartment: compartment_t = compartment_t.CELL,
    version: str = None,
    with_cell_labels: bool = True,
    with_track_labels: bool = True,
    mode: str = "2d+t",
    keep_every: int = 2,
    with_ticks: bool = True,
    prepare_only: bool = False,
    interactively: bool = True,
    in_main_thread: bool = True,
    dbe: context_t = DRAWING_CONTEXT,
) -> Optional[Union[figure_t, s_viewer_2d_t]]:
    """
    mode: see ShowSequence
    """
    output = None

    if mode == "2d+t":
        viewer = dbe.s_viewer_2d_t.NewForSegmentation(
            sequence,
            dbe,
            version=version,
            with_cell_labels=with_cell_labels,
            with_track_labels=with_track_labels,
            with_ticks=with_ticks,
        )
        if prepare_only:
            output = viewer
        else:
            viewer.Show(interactively=interactively, in_main_thread=in_main_thread)
    elif mode in ("mille-feuille", "tunnels"):
        if isinstance(sequence, (segmentations_t, sequence_t)):
            if isinstance(sequence, segmentations_t):
                segmentations = sequence
            else:
                segmentations = sequence.segmentations
            segmentations = segmentations.CompartmentsWithVersion(compartment)
        else:
            segmentations = sequence

        figure, axes = dbe.figure_3d_t.NewFigureAndAxes()
        axes: axes_3d_t
        if mode == "mille-feuille":
            if isinstance(sequence, sequence_t) and sequence.has_cells:
                cell_frames = sequence.cell_frames
            else:
                cell_frames = None
            tracks = gnrc.CellTracks(sequence, with_track_labels)
            gnr3.ShowFramesAsMilleFeuille(
                segmentations,
                None,
                with_cell_labels,
                cell_frames,
                tracks,
                axes,
                dbe.CellAnnotationStyle,
                keep_every=keep_every,
            )
        else:
            gnr3.ShowFramesAsTunnels(segmentations, axes, keep_every=keep_every)

        if prepare_only:
            output = figure
        else:
            figure.Show(interactively=interactively, in_main_thread=in_main_thread)
    else:
        raise ValueError(
            f'{mode}: Invalid mode; Expected="2d+t", "mille-feuille", "tunnels"'
        )

    return output



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

"""
This module was adapted from the module defining the function ``match_template`` of Scikit-Image.
Function documentation:
https://scikit-image.org/docs/dev/api/skimage.feature.html?highlight=match#skimage.feature.match_template
Module Code:
https://github.com/scikit-image/scikit-image/blob/main/skimage/feature/template.py

The original function ``match_template`` was adapted to the 1-D case, and the parameter ``normalized`` was added. Let us
remind that the Pearson correlation coefficient involves a normalization by the product of the standard deviations. See
for example:
https://en.wikipedia.org/wiki/Pearson_correlation_coefficient
The parameter ``normalized`` can take the following values:
    - None: no normalization is performed; the response is the local covariance; it is unbounded;
    - "max": the response is the local covariance C normalized by the maximum of C;
    - "max_abs": the response is the local covariance C normalized by the maximum of C in absolute value;
    - "stddev": corresponds to the original normalization, so that the response is the Pearson correlation
    coefficient.
    - "template": normalizes by numpy.sum((template - numpy.mean(template)) ** 2) so that the response to itself is 1.
"""

import warnings as wrng
from typing import Optional, Sequence, Union

import numpy as nmpy
from scipy.signal import fftconvolve


array_t = nmpy.ndarray


def _window_sum_1d(signal: array_t, window_length: int, /) -> array_t:
    """"""
    window_sum = nmpy.cumsum(signal)
    window_sum = window_sum[window_length:-1] - window_sum[: -window_length - 1]

    return window_sum


def match_template(
    signal: Union[array_t, Sequence[float]],
    template: Union[array_t, Sequence[float]],
    /,
    *,
    normalized: Optional[str] = None,
    pad_input: bool = False,
    mode: str = "constant",
    constant_values: float = 0.0,
) -> Optional[array_t]:
    """"""
    signal = nmpy.array(signal, dtype=nmpy.float64, copy=False)
    template = nmpy.array(template, dtype=nmpy.float64, copy=False)

    if (signal.ndim != template.ndim) or (signal.ndim != 1):
        raise ValueError(
            "Dimensionality of template must be "
            "equal to the dimensionality of signal. "
            "Both must be equal to one."
        )
    if signal.size < template.size:
        wrng.warn(
            f"{match_template.__name__}: Signal (length={signal.size}) "
            f"must be larger than template (length={template.size})"
        )
        return None

    signal_length = signal.size
    pad_length = template.size
    if mode == "constant":
        signal = nmpy.pad(
            signal, pad_width=pad_length, mode=mode, constant_values=constant_values
        )
    else:
        signal = nmpy.pad(signal, pad_width=pad_length, mode=mode)

    image_window_sum = _window_sum_1d(signal, template.size)

    template_mean = template.mean()
    template_length = template.size
    template_ssd = nmpy.sum((template - template_mean) ** 2)

    xcorr = fftconvolve(signal, template[::-1], mode="valid")[1:-1]

    numerator = xcorr - image_window_sum * template_mean

    if normalized is None:
        response = numerator
    elif normalized in ("max", "max_abs"):
        if normalized == "max":
            denominator = nmpy.amax(numerator)
        else:
            denominator = nmpy.amax(nmpy.fabs(numerator))
        if denominator > nmpy.finfo(numerator.dtype).eps:
            response = numerator / denominator
        else:
            response = numerator
    elif normalized == "stddev":
        image_window_sum2 = _window_sum_1d(signal ** 2, template.size)

        denominator = image_window_sum2
        nmpy.multiply(image_window_sum, image_window_sum, out=image_window_sum)
        nmpy.divide(image_window_sum, template_length, out=image_window_sum)
        denominator -= image_window_sum
        denominator *= template_ssd
        # sqrt of negative number not allowed
        nmpy.maximum(denominator, 0, out=denominator)
        nmpy.sqrt(denominator, out=denominator)

        response = nmpy.zeros_like(xcorr)

        # avoid zero-division
        mask = denominator > nmpy.finfo(response.dtype).eps

        response[mask] = numerator[mask] / denominator[mask]
    elif normalized == "template":
        response = numerator / template_ssd
    else:
        raise ValueError(f"{normalized}: Invalid normalization method")

    if pad_input:
        d0 = (template_length - 1) // 2
        d1 = d0 + signal_length
    else:
        d0 = template_length - 1
        d1 = d0 + signal_length - template_length + 1

    return response[slice(d0, d1)]



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from pathlib import Path as path_t
from typing import Callable, Sequence, Union

import numpy as nmpy
import tensorflow.keras.models as modl


array_t = nmpy.ndarray
processing_h = Callable[[array_t], array_t]


def SegmentationsWithTFNetwork(
    frames: Sequence[array_t],
    network_path: Union[str, path_t],
    /,
    *,
    threshold: float = 0.9,
    PreProcessed: processing_h = None,
    PostProcessed: processing_h = None,
) -> Sequence[array_t]:
    """
    PostProcessed: Could be used to clear border objects. However, since one might want to segment cytoplasms and
    nuclei, clearing border objects here could lead to clearing a cytoplasm while keeping its nucleus. Consequently,
    clearing border objects here, i.e. independently for each segmentation task, is not appropriate.
    """
    output = []

    if PreProcessed is not None:
        frames = tuple(PreProcessed(_frm) for _frm in frames)
    if PostProcessed is None:
        PostProcessed = lambda _prm: _prm

    frames = nmpy.array(frames, dtype=nmpy.float32)
    if frames.ndim == 3:
        frames = nmpy.expand_dims(frames, axis=3)

    network = modl.load_model(network_path)
    predictions = network.predict(frames, verbose=1)
    shape = network.layers[0].input_shape[0][1:-1]

    for t_idx, prediction in enumerate(predictions):
        reshaped = nmpy.reshape(prediction, shape)
        segmentation = reshaped > threshold
        post_processed = PostProcessed(segmentation)
        if nmpy.amax(post_processed.astype(nmpy.uint8)) == 0:
            raise ValueError(f"{t_idx}: Empty segmentation")

        output.append(post_processed)

    return output



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from typing import Sequence, Tuple

import numpy as nmpy
import scipy.ndimage as imge
import skimage.morphology as mrph
import skimage.transform as trsf


array_t = nmpy.ndarray


def InCommonNonZeroRectangles(
    reference: array_t,
    changed: array_t,
    /,
    *,
    as_squares: bool = False,
    for_rotation: bool = False,
    ensure_odd_sides: bool = False,
) -> Tuple[Sequence[array_t], Sequence[Sequence[int]]]:
    """"""
    rectangles = []
    corners = []

    tight_rectangles = []
    tight_shapes = []
    for image in (reference, changed):
        where = imge.find_objects(image > 0, max_label=1)[0]
        rectangle = image[where].copy()
        shape = nmpy.array(rectangle.shape)

        corners.append([_slc.start for _slc in where])
        tight_rectangles.append(rectangle)
        tight_shapes.append(shape)

    common_shape = nmpy.maximum(*tight_shapes)
    if for_rotation:
        side = int(nmpy.around(2.0 * nmpy.sqrt(nmpy.sum((0.5 * common_shape) ** 2))))
        common_shape = (side, side)
    elif as_squares:
        side = max(common_shape)
        common_shape = (side, side)
    if ensure_odd_sides:
        common_shape += 1 - (common_shape % 2)
    common_shape = nmpy.array(common_shape)

    for corner, shape, rectangle in zip(corners, tight_shapes, tight_rectangles):
        if nmpy.any(shape < common_shape):
            common_rectangle = nmpy.zeros(common_shape, dtype=rectangle.dtype)
            differences = (common_shape - shape) // 2
            common_rectangle[
                differences[0] : (differences[0] + shape[0]),
                differences[1] : (differences[1] + shape[1]),
            ] = rectangle
            rectangles.append(common_rectangle)
            for c_idx in range(2):
                corner[c_idx] -= differences[c_idx]
        else:
            rectangles.append(rectangle)

    return rectangles, corners


def RotationInBinary(reference: array_t, rotated: array_t, /) -> float:
    """"""
    output = 0.0

    distance_map = imge.distance_transform_edt(reference)
    highest_distance = 0.0
    for angle in nmpy.linspace(-20.0, 20.0, num=11):
        attempt = trsf.rotate(rotated, angle, preserve_range=True)
        distance = nmpy.sum(distance_map * attempt)
        if distance > highest_distance:
            highest_distance = distance
            output = angle

    return output


def RotatedLabeled(labeled: array_t, angle: float, /) -> array_t:
    """"""
    output = nmpy.zeros_like(labeled)

    rotated = trsf.rotate(labeled, angle, preserve_range=True)
    n_labels = nmpy.amax(labeled)
    for label in range(1, n_labels + 1):
        where = nmpy.logical_and(rotated >= label - 0.5, rotated < label + 0.5)
        sublabeled, n_sublabels = mrph.label(where, return_num=True, connectivity=1)

        highest_area = 0
        best_sublabel = None
        for sublabel in range(1, n_sublabels + 1):
            area = nmpy.count_nonzero(sublabeled == sublabel)
            if area > highest_area:
                highest_area = area
                best_sublabel = sublabel
        output[sublabeled == best_sublabel] = label

    return output



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from matplotlib.artist import Artist as artist_t

import cell_tracking_BC.in_out.graphics.dbe.matplotlib.any_d as anyd
import cell_tracking_BC.in_out.graphics.dbe.matplotlib.style as styl
import cell_tracking_BC.in_out.graphics.dbe.matplotlib.three_d as thrd
import cell_tracking_BC.in_out.graphics.dbe.matplotlib.two_d as twod
from cell_tracking_BC.in_out.graphics.context import context_t


DRAWING_CONTEXT = context_t(
    figure_2d_t=twod.figure_t,
    axes_2d_t=twod.axes_t,
    s_viewer_2d_t=twod.s_viewer_2d_t,
    t_viewer_2d_t=twod.t_viewer_2d_t,
    #
    figure_3d_t=thrd.figure_t,
    axes_3d_t=thrd.axes_t,
    #
    CellAnnotationStyle=styl.CellAnnotationStyle,
    #
    NewSlider=anyd.NewSlider,
    SliderValue=anyd.SliderValue,
    UpdateSlider=anyd.UpdateSlider,
    SliderBounds=lambda _sld: (_sld.valmin, _sld.valmax),
    SliderAxes=lambda _sld: _sld.ax,
    #
    IsVisible=artist_t.get_visible,
    SetVisibility=artist_t.set_visible,
    #
    # _evt: matplotlib.backend_bases.LocationEvent
    IsTargetOfEvent=lambda _axs, _evt: _evt.inaxes is _axs,
    KeyEventKey=lambda _evt: _evt.key,
    ScrollEventStep=lambda _evt: _evt.step,
)



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from __future__ import annotations

import dataclasses as dcls
from typing import Any, Dict, Optional, Sequence, Tuple

import numpy as nmpy
import scipy.ndimage as spim
import skimage.feature as ftre
import pca_b_stream as pcst
from PIL import Image as pil_image_t

from cell_tracking_BC.standard.uid import ShortID


array_t = nmpy.ndarray


@dcls.dataclass(repr=False, eq=False)
class compartment_t:

    centroid: array_t = None  # In frame coordinates
    bb_slices: Tuple[slice, slice] = None  # bb=bounding box; In frame coordinates
    touches_border: bool = None
    map_stream: Optional[bytes] = None  # Built from a numpy.bool_ array
    features: Dict[str, Any] = dcls.field(init=False, default_factory=dict)

    @classmethod
    def NewFromMap(cls, map_: array_t, /, *, is_plain: bool = True) -> compartment_t:
        """"""
        if is_plain:
            # By default, ftre.peak_local_max excludes peaks on the frame border. This can be parameterized otherwise,
            # but this would lead to an "off-center" peak anyway. Instead, the compartment map is padded.
            padded = nmpy.pad(map_, 1)
            distance_map = spim.distance_transform_edt(padded)
            centroid = ftre.peak_local_max(distance_map, num_peaks=1)[0] - 1
        else:
            pixels = nmpy.nonzero(map_)
            centroid_as_it = (
                nmpy.mean(pixels[_idx]) for _idx in range(pixels.__len__())
            )
            centroid = nmpy.fromiter(centroid_as_it, dtype=nmpy.float64)

        map_as_pil = pil_image_t.fromarray(map_)
        auto_crop_lengths = map_as_pil.getbbox()
        # auto_crop_lengths[2 and 3] already include the +1 for slices
        bb_slices = (
            slice(auto_crop_lengths[1], auto_crop_lengths[3]),
            slice(auto_crop_lengths[0], auto_crop_lengths[2]),
        )
        touches_border = (
            (auto_crop_lengths[0] == 0)
            or (auto_crop_lengths[1] == 0)
            or (auto_crop_lengths[2] == map_.shape[1])
            or (auto_crop_lengths[3] == map_.shape[0])
        )

        cropped = map_as_pil.crop(auto_crop_lengths)
        # TODO: Changing bool_ to int8 or uint8 breaks map generation (To be investigated one day)
        map_stream = pcst.PCA2BStream(nmpy.array(cropped, dtype=nmpy.bool_))

        instance = cls(
            centroid=centroid,
            bb_slices=bb_slices,
            touches_border=touches_border,
            map_stream=map_stream,
        )

        return instance

    def AddFeature(self, name: str, value: Any, /) -> None:
        """"""
        self.features[name] = value

    def AddFeaturesFromDictionary(self, features: Dict[str, Any], /) -> None:
        """"""
        self.features.update(features)

    @property
    def available_features(self) -> Sequence[str]:
        """"""
        return sorted(self.features.keys())

    def Map(self, shape: Sequence[int], /, *, as_boolean: bool = False) -> array_t:
        """"""
        if as_boolean:
            dtype = nmpy.bool_
            one = True
        else:
            dtype = nmpy.int8  # Do not use uint8 to allow map subtraction
            one = 1
        output = nmpy.zeros(shape, dtype=dtype)

        bb_map = self.BBMap()
        output[self.bb_slices][bb_map] = one

        return output

    def BBMap(self) -> array_t:
        """
        BB=Bounding box=Map limited to the compartment bounding box. Dtype: numpy.bool_
        """
        return pcst.BStream2PCA(self.map_stream)

    def NonZeroPixels(self, shape: Sequence[int], /) -> Tuple[array_t, array_t]:
        """"""
        return nmpy.nonzero(self.Map(shape))

    def __str__(self) -> str:
        """"""
        return (
            f"{self.__class__.__name__.upper()}.{ShortID(id(self))}:\n"
            f"    {self.centroid=}\n"
            f"    {self.bb_slices=}\n"
            f"    {self.map_stream=}\n"
            f"    {self.features=}"
        )



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from typing import Any, Callable, Optional, Sequence, Tuple, Union

import numpy as nmpy
import skimage.measure as msre
from cell_tracking_BC.type.cell import cell_t


array_t = nmpy.ndarray


SKIMAGE_MORPHOLOGICAL_FEATURES: Tuple[str]  # Set below
SKIMAGE_RADIOMETRIC_FEATURES: Tuple[str]  # Set below
SKIMAGE_ALL_FEATURES: Tuple[str]  # Set below


def NumpyArrayScalarFeatureInCell(
    cell: cell_t, frame: array_t, Feature: Callable[[array_t], array_t], /
) -> Union[int, float]:
    """
    Feature: the array_t returned value must be an array with a single value
    """
    output = NumpyArrayAnyFeatureInCell(cell, frame, Feature)

    return output.item()


def NumpyArrayAnyFeatureInCell(
    cell: cell_t, frame: array_t, Feature: Callable[[array_t], array_t], /
) -> array_t:
    """"""
    map_ = cell.Map(frame.shape)
    output = Feature(frame[map_])

    return output


def CellSKImageMorphologicalFeatures(cell: cell_t, /) -> Sequence[Any]:
    """"""
    return _CellSKImageFeatures(cell, None, SKIMAGE_MORPHOLOGICAL_FEATURES)


def CellSKImageAllFeatures(cell: cell_t, frame: array_t, /) -> Sequence[Any]:
    """"""
    return _CellSKImageFeatures(
        cell, frame, SKIMAGE_MORPHOLOGICAL_FEATURES + SKIMAGE_RADIOMETRIC_FEATURES
    )


def _CellSKImageFeatures(
    cell: cell_t, frame: Optional[array_t], names: Sequence[str], /
) -> Sequence[Any]:
    """"""
    output = []

    if frame is None:
        # /!\ Because the bounding box map is used (as opposed to the full, frame-shaped map), features such as the
        # centroid are relative to the cell bounding box. However, features related to absolute positioning are usually
        # not used for clustering or classification.
        bb_map = cell.BBMap().astype(nmpy.int8)
        features = msre.regionprops(bb_map)[0]
    else:
        cell_map = cell.Map(frame.shape)
        features = msre.regionprops(cell_map, intensity_image=frame)[0]
    for name in names:
        output.append(features[name])

    return output


def CellArea(cell: cell_t, /) -> int:
    """"""
    bb_map = cell.BBMap()
    output = nmpy.count_nonzero(bb_map)

    return output


def IntensityEntropyInCell(
    cell: cell_t,
    frame: array_t,
    /,
    *,
    n_bins: int = None,
    normalized_version: bool = True,
) -> float:
    """
    normalized_version: It True, the entropy is normalized by 1/log(n_bins) so that the output belongs to [0,1]
    """
    map_ = cell.Map(frame.shape)
    intensities = frame[map_]

    if n_bins is None:
        n_bins = max(3, int(round(nmpy.sqrt(intensities.size))))
    histogram, _ = nmpy.histogram(intensities, bins=n_bins)
    normed = histogram / nmpy.sum(histogram)
    non_zero = normed[normed > 0.0]

    output = -nmpy.sum(non_zero * nmpy.log(non_zero))
    if normalized_version:
        output /= nmpy.log(n_bins)

    return output.item()


def _CellSKImageMorphologicalFeatureNames() -> Tuple[str]:
    """"""
    return _CellSKImageFeatureNames(False)


def _CellSKImageAllFeatureNames() -> Tuple[str]:
    """"""
    return _CellSKImageFeatureNames(True)


def _CellSKImageFeatureNames(with_radiometry: bool, /) -> Tuple[str]:
    """"""
    output = []

    dummy = nmpy.zeros((10, 10), dtype=nmpy.int8)
    dummy[4:7, 4:7] = 1
    if with_radiometry:
        features = msre.regionprops(dummy, intensity_image=dummy)[0]
    else:
        features = msre.regionprops(dummy)[0]

    for name in dir(features):
        if (not name.startswith("_")) and hasattr(features, name):
            output.append(name)

    return tuple(output)


SKIMAGE_MORPHOLOGICAL_FEATURES = _CellSKImageMorphologicalFeatureNames()
SKIMAGE_RADIOMETRIC_FEATURES = tuple(
    set(_CellSKImageAllFeatureNames()).difference(SKIMAGE_MORPHOLOGICAL_FEATURES)
)
# Keep this sum instead of _CellSKImageAllFeatureNames() to guaranty the correct order
SKIMAGE_ALL_FEATURES = SKIMAGE_MORPHOLOGICAL_FEATURES + SKIMAGE_RADIOMETRIC_FEATURES



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from __future__ import annotations

import dataclasses as dtcl
import datetime as dttm
import re as regx
from pathlib import Path as path_t
from typing import Callable, Dict, Optional, Sequence, Tuple, Union

import numpy as nmpy
import tifffile as tiff

import cell_tracking_BC.in_out.graphics.generic.any_d as gphc
from cell_tracking_BC.in_out.graphics.context import _sg, annotation_h
from cell_tracking_BC.in_out.graphics.context import axes_2d_t as axes_t
from cell_tracking_BC.in_out.graphics.context import (
    context_t,
    event_h,
    figure_t,
    key_event_h,
    scroll_event_h,
    slider_h,
)
from cell_tracking_BC.type.frame import frame_t
from cell_tracking_BC.type.segmentation import compartment_t, segmentation_t
from cell_tracking_BC.type.segmentations import segmentations_t
from cell_tracking_BC.type.sequence import BoundingBoxSlices, sequence_h, sequence_t
from cell_tracking_BC.type.tracks import tracks_t


array_t = nmpy.ndarray

all_versions_h = Dict[
    str, Tuple[Tuple[int, int], Union[Sequence[array_t], Sequence[frame_t]]]
]


@dtcl.dataclass(repr=False, eq=False)
class s_viewer_2d_t:

    figure: figure_t
    axes: axes_t
    slider: Optional[slider_h]

    all_versions: all_versions_h
    current_version: str = None
    current_time_point: int = -1  # Used only when slider is None
    current_label: int = -1

    # Only meaningful for NewForChannels
    cell_contours: Sequence[Sequence[array_t]] = None
    # Using frame labeling for array_t's, or cell_frames below for sequence_t
    with_cell_labels: bool = False
    cell_frames: Sequence[frame_t] = None
    tracks: tracks_t = None
    annotations: Sequence[Tuple[int, Union[annotation_h, Sequence[annotation_h]]]] = None

    dbe: context_t = None

    @classmethod
    def NewForChannels(
        cls,
        sequence: Union[Sequence[array_t], sequence_t],
        dbe: context_t,
        /,
        *,
        channel: str = None,
        with_segmentation: bool = False,
        with_cell_labels: bool = False,
        with_track_labels: bool = False,
        in_axes: axes_t = None,
        with_ticks: bool = True,
        with_colorbar: bool = True,
    ) -> s_viewer_2d_t:
        """"""
        with_cell_labels = (
            with_cell_labels and isinstance(sequence, sequence_t) and sequence.has_cells
        )
        if isinstance(sequence, sequence_t) and sequence.has_cells:
            cell_frames = sequence.cell_frames
        else:
            cell_frames = None

        instance = cls._NewForSequence(
            sequence,
            _AllChannelsOfSequence,
            dbe,
            version=channel,
            with_segmentation=with_segmentation,
            with_cell_labels=with_cell_labels,
            cell_frames=cell_frames,
            with_track_labels=with_track_labels,
            in_axes=in_axes,
            with_ticks=with_ticks,
        )
        if with_colorbar:
            instance.AddColorbarForImage()

        return instance

    @classmethod
    def NewForSegmentation(
        cls,
        sequence: sequence_h,
        dbe: context_t,
        /,
        *,
        version: str = None,
        with_cell_labels: bool = True,
        with_track_labels: bool = True,
        in_axes: axes_t = None,
        with_ticks: bool = True,
    ) -> s_viewer_2d_t:
        """"""
        if (
            isinstance(sequence, Sequence)
            and isinstance(sequence[0], segmentation_t)
            and not isinstance(sequence, segmentations_t)
        ):
            new_sequence = segmentations_t()
            for segmentation in sequence:
                new_sequence.append(segmentation)
            sequence = new_sequence

        if isinstance(sequence, sequence_t) and sequence.has_cells:
            cell_frames = sequence.cell_frames
        else:
            cell_frames = None

        return cls._NewForSequence(
            sequence,
            _AllSegmentationsOfSequence,
            dbe,
            version=version,
            with_cell_labels=with_cell_labels,
            cell_frames=cell_frames,
            with_track_labels=with_track_labels,
            in_axes=in_axes,
            with_ticks=with_ticks,
        )

    @classmethod
    def NewForAllStreams(
        cls,
        sequence: sequence_t,
        dbe: context_t,
        /,
        *,
        version: str = None,
        with_segmentation: bool = False,
        with_cell_labels: bool = False,
        with_track_labels: bool = False,
        in_axes: axes_t = None,
        with_ticks: bool = True,
        with_colorbar: bool = True,
    ) -> s_viewer_2d_t:
        """"""
        with_cell_labels = with_cell_labels and sequence.has_cells
        if sequence.has_cells:
            cell_frames = sequence.cell_frames
        else:
            cell_frames = None

        instance = cls._NewForSequence(
            sequence,
            _AllStreamsOfSequence,
            dbe,
            version=version,
            with_segmentation=with_segmentation,
            with_cell_labels=with_cell_labels,
            cell_frames=cell_frames,
            with_track_labels=with_track_labels,
            in_axes=in_axes,
            with_ticks=with_ticks,
        )
        if with_colorbar:
            instance.AddColorbarForImage()

        return instance

    @classmethod
    def _NewForSequence(
        cls,
        sequence: sequence_h,
        AllVersionsOfSequence: Callable[[sequence_h], Tuple[all_versions_h, str]],
        dbe: context_t,
        /,
        *,
        version: str = None,
        with_segmentation: bool = False,
        with_cell_labels: bool = True,
        cell_frames: Sequence[frame_t] = None,
        with_track_labels: bool = True,
        in_axes: axes_t = None,
        with_ticks: bool = True,
    ) -> s_viewer_2d_t:
        """"""
        if in_axes is None:
            figure, axes = dbe.figure_2d_t.NewFigureAndAxes()
        else:
            figure = in_axes.Figure()
            axes = in_axes
        if not with_ticks:
            axes.TurnTicksOff()

        all_versions, current_version = AllVersionsOfSequence(sequence)
        if version is not None:
            current_version = version
        if more_than_one := (all_versions.__len__() > 1):
            axes.SetTitle(current_version)

        cell_contours = gphc.CellContours(sequence, with_segmentation)
        tracks = gphc.CellTracks(sequence, with_track_labels)

        n_frames, cell_annotations = _ShowFirstFrame(
            all_versions,
            current_version,
            cell_contours,
            with_cell_labels,
            cell_frames,
            tracks,
            axes,
            dbe.CellAnnotationStyle,
        )

        if in_axes is None:
            slider = dbe.NewSlider(figure, n_frames)
            current_time_point = None
        else:
            slider = None
            current_time_point = 0

        instance = cls(
            figure=figure,
            axes=axes,
            slider=slider,
            all_versions=all_versions,
            current_version=current_version,
            current_time_point=current_time_point,
            cell_contours=cell_contours,
            with_cell_labels=with_cell_labels,
            tracks=tracks,
            cell_frames=cell_frames,
            annotations=cell_annotations,
            dbe=dbe,
        )
        instance._ActivateEventProcessing(more_than_one)

        return instance

    def AddColorbarForImage(self) -> None:
        """"""
        raise NotImplementedError

    def _ActivateEventProcessing(self, more_than_one_version: bool, /) -> None:
        """"""
        self.figure.ActivateEvent("key_press_event", self._OnKeyPress)
        if more_than_one_version:
            self.figure.ActivateEvent("button_press_event", self._OnButtonPress)
        if self.slider is not None:
            self.figure.ActivateEvent("scroll_event", self._OnScrollEvent)

    def ShowFrame(
        self,
        /,
        *,
        version: str = None,
        time_point: Union[int, float] = None,
        highlighted: int = -1,
        should_update_limits: bool = False,
        should_draw: bool = True,
        force_update: bool = False,
    ) -> None:
        """
        force_update: If the slider has been updated externally, the time point will not be considered new, and no
        update will be made. Hence, this parameter.
        """
        if version is None:
            version = self.current_version
        if self.slider is None:
            current_time_point = self.current_time_point
        else:
            current_time_point = self.slider.val
        if time_point is None:
            time_point = current_time_point
        else:
            time_point = int(time_point)

        version_is_new = version != self.current_version
        time_point_is_new = (time_point != current_time_point) or force_update

        if version_is_new or time_point_is_new:
            interval, frames = self.all_versions[version]
            frame = frames[time_point]
            self.axes.UpdateImage(
                frame, interval=interval, should_update_limits=should_update_limits
            )
        else:
            frame = None

        if self.annotations is not None:
            if time_point_is_new:
                if self.cell_contours is None:
                    contours = None
                else:
                    contours = self.cell_contours[time_point]
                if self.cell_frames is None:
                    cell_frame = None
                else:
                    cell_frame = self.cell_frames[time_point]

                self.annotations = gphc.AnnotateCells(
                    frame,
                    contours,
                    self.with_cell_labels,
                    cell_frame,
                    self.tracks,
                    self.axes,
                    self.dbe.CellAnnotationStyle,
                    highlighted=highlighted,
                )
                self.current_label = highlighted
            elif highlighted > 0:
                self.HighlightAnnotation(highlighted, should_draw=False)

        if version_is_new:
            self.axes.SetTitle(version)
            self.current_version = version

        if time_point_is_new:
            if self.slider is None:
                self.current_time_point = time_point
            else:
                self.dbe.UpdateSlider(self.slider, time_point)

        if should_draw:
            self.figure.Update()

    def ShowNextVersion(self, /, *, should_draw: bool = True) -> None:
        """"""
        all_names = tuple(self.all_versions.keys())
        where = all_names.index(self.current_version)
        where = (where + 1) % all_names.__len__()
        new_version = all_names[where]

        self.ShowFrame(
            version=new_version, should_update_limits=True, should_draw=should_draw
        )

    def HighlightAnnotation(self, label: int, /, *, should_draw: bool = True) -> None:
        """
        If label is <= 0 or > max cell label in current frame, then un-highlights all annotations
        """
        raise NotImplementedError

    def Show(
        self,
        /,
        *,
        interactively: bool = True,
        in_main_thread: bool = True,
    ) -> None:
        """"""
        self.figure.Show(interactively=interactively, in_main_thread=in_main_thread)

    def _OnKeyPress(
        self,
        event: key_event_h,
        /,
    ) -> None:
        """"""
        if self.dbe.KeyEventKey(event).lower() == "s":
            print("Sequence saving in progress...")
            volume = self.AsAnnotatedVolume()

            illegal = "[^-_a-zA-Z0-9]"
            version = regx.sub(illegal, "", self.current_version)
            now = regx.sub(illegal, "-", dttm.datetime.now().isoformat())
            path = path_t.home() / f"sequence-{version}-{now}.tif"
            if path.exists():
                print(f"{path}: Existing path; Cannot override")
                return

            tiff.imwrite(
                str(path),
                volume,
                photometric="rgb",
                compression="deflate",
                planarconfig="separate",
                metadata={"axes": "XYZCT"},
            )
            print(f"Annotated sequence saved at: {path}")

    def _OnButtonPress(
        self,
        event: event_h,
        /,
    ) -> None:
        """"""
        if self.dbe.IsTargetOfEvent(self.axes, event):
            self.ShowNextVersion()
        elif (self.slider is not None) and self.dbe.IsTargetOfEvent(
            self.dbe.SliderAxes(self.slider), event
        ):
            self.ShowFrame(
                time_point=self.dbe.SliderValue(self.slider), force_update=True
            )

    def _OnScrollEvent(self, event: scroll_event_h) -> None:
        """"""
        value = self.dbe.SliderValue(self.slider)
        bounds = self.dbe.SliderBounds(self.slider)
        new_value = round(value + nmpy.sign(self.dbe.ScrollEventStep(event)))
        new_value = min(max(new_value, bounds[0]), bounds[1])
        if new_value != value:
            self.ShowFrame(time_point=new_value)

    def AsAnnotatedVolume(self) -> array_t:
        """
        See: cell_tracking_BC.in_out.file.sequence.SaveAnnotatedSequence
        """
        # Cannot be initialized since content (not frame) shape is unknown
        output = None

        current_version = self.all_versions[self.current_version][1]
        sequence_length = current_version.__len__()

        figure, axes = self.dbe.figure_2d_t.NewFigureAndAxes()
        axes: axes_t
        invisible = self.__class__._NewForSequence(
            current_version,
            _AllChannelsOfSequence,
            self.dbe,
            with_cell_labels=self.with_cell_labels,
            cell_frames=self.cell_frames,
            with_track_labels=False,
            with_ticks=False,
            in_axes=axes,
        )
        invisible.cell_contours = self.cell_contours
        invisible.tracks = self.tracks

        for time_point in range(sequence_length):
            invisible.ShowFrame(time_point=time_point, should_draw=False)
            # draw_idle is not appropriate here
            figure.Update(gently=False)
            content = figure.Content()
            if output is None:
                output = nmpy.empty((*content.shape, sequence_length), dtype=nmpy.uint8)
            output[..., time_point] = content

        figure.Close()  # To prevent remaining caught in event loop

        row_slice, col_slice = BoundingBoxSlices(output)
        output = output[row_slice, col_slice, :, :]
        output = nmpy.moveaxis(output, (0, 1, 2, 3), (2, 3, 1, 0))
        output = output[:, nmpy.newaxis, :, :, :]

        return output


def _AllChannelsOfSequence(
    sequence: Union[Sequence[array_t], sequence_t]
) -> Tuple[all_versions_h, str]:
    """"""
    if isinstance(sequence, sequence_t):
        all_channels = {}
        for channel in sequence.channels:
            frames = sequence.Frames(channel=channel)
            min_value = min(nmpy.amin(_frm) for _frm in frames)
            max_value = max(nmpy.amax(_frm) for _frm in frames)
            all_channels[channel] = ((min_value, max_value), frames)
        current_channel = sequence.channels[0]
    else:
        current_channel = "MAIN"
        min_value = min(nmpy.amin(_frm) for _frm in sequence)
        max_value = max(nmpy.amax(_frm) for _frm in sequence)
        all_channels = {current_channel: ((min_value, max_value), sequence)}

    return all_channels, current_channel


def _AllSegmentationsOfSequence(sequence: sequence_h) -> Tuple[all_versions_h, str]:
    """"""
    if isinstance(sequence, (segmentations_t, sequence_t)):
        if isinstance(sequence, segmentations_t):
            segmentations = sequence
        else:
            segmentations = sequence.segmentations

        all_versions = {}
        compartments, versions = segmentations.available_versions
        for compartment in compartments:
            for version in versions:
                key = f"{compartment.name}:{version[0]}:{version[1]}"
                frames = segmentations.CompartmentsWithVersion(
                    compartment, index=version[0], name=version[1]
                )
                all_versions[key] = ((0, 1), frames)
        current_version = f"{compartment_t.CELL.name}:{versions[0][0]}:{versions[0][1]}"
    else:
        current_version = "MAIN"
        all_versions = {current_version: ((0, 1), sequence)}

    return all_versions, current_version


def _AllStreamsOfSequence(sequence: sequence_t) -> Tuple[all_versions_h, str]:
    """"""
    all_streams, current_stream = _AllChannelsOfSequence(sequence)
    all_versions, _ = _AllSegmentationsOfSequence(sequence)

    all_streams.update(all_versions)

    return all_streams, current_stream


def _ShowFirstFrame(
    all_versions: all_versions_h,
    current_version: str,
    cell_contours: Optional[Sequence[Sequence[array_t]]],
    with_cell_labels: bool,
    cell_frames: Optional[Sequence[frame_t]],
    tracks: Optional[tracks_t],
    axes: axes_t,
    CellAnnotationStyle: _sg.cell_annotation_style_h,
    /,
) -> Tuple[int, Optional[Sequence[Tuple[int, annotation_h]]]]:
    """"""
    interval, version = all_versions[current_version]
    first_frame = version[0]

    axes.PlotImage(first_frame, interval=interval)

    if (cell_contours is not None) or with_cell_labels or (tracks is not None):
        if cell_contours is None:
            contours = None
        else:
            contours = cell_contours[0]
        if cell_frames is None:
            cell_frame = None
        else:
            cell_frame = cell_frames[0]
        cell_annotations = gphc.AnnotateCells(
            first_frame,
            contours,
            with_cell_labels,
            cell_frame,
            tracks,
            axes,
            CellAnnotationStyle,
        )
    else:
        cell_annotations = None

    # Once the first frame has been plot, disable axes autoscale to try to speed future plots up
    axes.Freeze()

    return version.__len__(), cell_annotations



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from __future__ import annotations

import dataclasses as dcls
from enum import Enum as enum_t
from typing import Any, Optional, Sequence, Tuple, Union

import numpy as nmpy
import scipy.ndimage.morphology as scph

from cell_tracking_BC.type.compartment import compartment_t
from cell_tracking_BC.type.cytoplasm import cytoplasm_t
from cell_tracking_BC.type.nucleus import nucleus_t


array_t = nmpy.ndarray
nucleus_or_nuclei_h = Union[nucleus_t, Tuple[nucleus_t, nucleus_t]]


@dcls.dataclass(repr=False, eq=False)
class state_t(enum_t):
    unknown = None
    pruned = -1
    living = 1
    dividing = 2
    dead = 0

    def IsActive(self) -> bool:
        """"""
        return self not in (state_t.dead, state_t.pruned)


@dcls.dataclass(repr=False, eq=False)
class cell_t(compartment_t):

    label: Any = None
    nucleus: Optional[nucleus_or_nuclei_h] = None
    cytoplasm: Optional[cytoplasm_t] = None
    # Inherited from compartment_t (among others): features
    state: state_t = state_t.unknown

    @classmethod
    def NewFromMap(cls, _: array_t, /, *, is_plain: bool = True) -> compartment_t:
        """"""
        raise RuntimeError(
            f"{cell_t.NewFromMap.__name__}: Not meant to be called from class {cell_t.__name__}; "
            f"Use {cell_t.NewFromMaps.__name__} instead"
        )

    @classmethod
    def NewFromMaps(
        cls,
        label: Any,
        /,
        *,
        cell_map: array_t = None,
        cytoplasm_map: array_t = None,
        nucleus_map: array_t = None,
    ) -> cell_t:
        """
        cell_map: Defined optional to follow cytoplasm and nucleus, but never None in fact
        """
        # TODO: make this implementation coherent with segmentation_t, and check the Map method. In particular, is it
        #       necessary to have nuclei and the associated management when the cytoplasm holes play this role already?
        #       Also, make clear in the doc that the 3 maps are supposed to be coherent since all the checks have been
        #       done in segmentation_t (at least should have been).
        cell = compartment_t.NewFromMap(cell_map)
        if cytoplasm_map is None:
            cytoplasm = None
        else:
            cytoplasm = cytoplasm_t.NewFromMap(cytoplasm_map, is_plain=False)
        if nucleus_map is None:
            nucleus = None
        else:
            nucleus = nucleus_t.NewFromMap(nucleus_map)

        instance = cls(
            label=label,
            nucleus=nucleus,
            cytoplasm=cytoplasm,
            centroid=cell.centroid,
            bb_slices=cell.bb_slices,
            touches_border=cell.touches_border,
            map_stream=cell.map_stream,
        )

        return instance

    def AddNucleus(self, nucleus_map: array_t, /) -> None:
        """
        For dividing cells
        """
        if self.nucleus is None:
            raise RuntimeError(f"Cannot add nucleus to a cell without a nucleus yet")

        nucleus = nucleus_t.NewFromMap(nucleus_map)
        self.nucleus = (self.nucleus, nucleus)

    @property
    def nuclei(self) -> Tuple[nucleus_t, ...]:
        """"""
        if self.nucleus is None:
            output = ()
        elif isinstance(self.nucleus, nucleus_t):
            output = (self.nucleus,)
        else:
            output = self.nucleus

        return output

    def Map(
        self,
        shape: Sequence[int],
        /,
        *,
        as_boolean: bool = False,
        with_labels: bool = False,
    ) -> array_t:
        """
        with_labels: cytoplasm will be marked with label 1, nuclei with subsequent labels
        """
        if as_boolean:
            dtype = nmpy.bool_
            one = True
        else:
            dtype = nmpy.uint8
            one = 1
        output = nmpy.zeros(shape, dtype=dtype)

        if self.cytoplasm is None:
            outer_compartment = self
        else:
            outer_compartment = self.cytoplasm
        nuclei = self.nuclei
        for c_idx, compartment in enumerate((outer_compartment, *nuclei), start=1):
            if compartment is self:
                map_ = super().Map(shape, as_boolean=True)
            else:
                map_ = compartment.Map(shape, as_boolean=True)
            if with_labels:
                output[map_] = c_idx
            else:
                output[map_] = one
        if nuclei.__len__() == 0:
            # TODO: this should not be necessary since the cell should represent the whole shape already
            output = scph.binary_fill_holes(output).astype(dtype, copy=False)

        return output

    def __str__(self) -> str:
        """"""
        super_str = super().__str__()

        lines = [super_str, "--- With compartments:"]
        initial_n_lines = lines.__len__()
        for compartment in (*self.nuclei, self.cytoplasm):
            if compartment is not None:
                lines.append(str(compartment))
        if lines.__len__() == initial_n_lines:
            return super_str

        return "\n".join(lines)



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import shelve as shlv
import sys as sstm
from pathlib import Path as path_t
from typing import Any, Union


def CacheDictionaryForMain(
    main: Union[str, path_t],
    /,
    *,
    folder_name: str = "_runtime",
    folder_postfix: str = "",
    cache_postfix: str = "",
) -> shlv.Shelf:
    """"""
    if isinstance(main, str):
        main = path_t(main)

    cache_folder = main.parent / (folder_name + folder_postfix)
    if not cache_folder.exists():
        cache_folder.mkdir()
    elif not cache_folder.is_dir():
        print(f"{cache_folder}: Not a suitable cache folder")
        sstm.exit(-1)

    return shlv.open(str(cache_folder / (main.stem + cache_postfix)))


def CloseCache(cache: Union[shlv.Shelf, Any]) -> None:
    """"""
    if isinstance(cache, shlv.Shelf):
        cache.close()



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import dataclasses as dcls
import logging as lggg
from typing import ClassVar

from rich.color import Color as color_t
from rich.highlighter import Highlighter as base_highlighter_t
from rich.logging import RichHandler as rich_handler_t
from rich.style import Style as style_t
from rich.text import Text as text_t


@dcls.dataclass(init=False, repr=False, eq=False)
class highlighter_t(base_highlighter_t):
    _WHERE: ClassVar[str] = r"[^:]+:[^:]+:[0-9]+"
    _ACTUAL: ClassVar[str] = r" Actual=[^.]+\."
    _EXPECTED: ClassVar[str] = r" Expected([!<>]=|: )[^.]+\."
    _GRAY_STYLE: ClassVar[style_t] = style_t(color=color_t.from_rgb(150, 150, 150))

    def highlight(self, text: text_t) -> None:
        """"""
        if text.plain[-1] != ".":
            text.append(".")
        text.highlight_regex(self.__class__._WHERE, style=self.__class__._GRAY_STYLE)
        text.highlight_regex(self.__class__._ACTUAL, style="red")
        text.highlight_regex(self.__class__._EXPECTED, style="green")


_HANDLER = rich_handler_t(
    highlighter=highlighter_t(),
    omit_repeated_times=False,
    enable_link_path=False,
    show_path=False,
)
_HANDLER.KEYWORDS = None
lggg.basicConfig(
    level="INFO",
    format="%(module)s:%(funcName)s:%(lineno)d %(message)s",
    datefmt="%Y-%m-%d@%H:%M:%S",
    handlers=[_HANDLER],
)

LOGGER = lggg.getLogger("rich")


def WhereFunction(function: object, /) -> str:
    """"""
    return f"{function.__module__}:{function.__name__}"


def WhereMethod(obj: object, method: object, /) -> str:
    """
    method: Could be a str instead, which would require changing method.__name__ into getattr(cls, method). But if the
        method name changes while forgetting to change the string in the call to WhereMethod accordingly, then an
        exception would be raised here.
    """
    cls = obj.__class__

    return f"{cls.__module__}:{cls.__name__}:{method.__name__}"



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from typing import Optional, Sequence

import numpy as nmpy

from cell_tracking_BC.in_out.graphics.context import _sg
from cell_tracking_BC.in_out.graphics.context import axes_3d_t as axes_t
from cell_tracking_BC.in_out.graphics.generic.any_d import AnnotateCells
from cell_tracking_BC.type.frame import frame_t
from cell_tracking_BC.type.tracks import tracks_t


array_t = nmpy.ndarray


_MILLEFEUILLE_ALPHA = 0.8


def ShowFramesAsMilleFeuille(
    frames: Sequence[array_t],
    cell_contours: Optional[Sequence[Sequence[array_t]]],
    with_cell_labels: bool,
    cell_frames: Optional[Sequence[frame_t]],
    tracks: Optional[tracks_t],
    axes: axes_t,
    CellAnnotationStyle: _sg.cell_annotation_style_h,
    /,
    *,
    keep_every: int = 2,
    n_levels: int = 1,
) -> None:
    """"""
    n_frames = frames.__len__()
    shape = frames[0].shape
    scaling = axes.__class__.MillefeuilleScaling(shape, n_frames)

    kept_frames = frames[::keep_every]
    min_intensity = min(nmpy.amin(_frm) for _frm in kept_frames)
    max_intensity = max(nmpy.amax(_frm) for _frm in kept_frames)
    intensity_range = max(1, max_intensity - min_intensity)

    all_rows, all_cols = nmpy.meshgrid(range(shape[0]), range(shape[1]), indexing="ij")
    for t_idx, frame in enumerate(frames):
        if t_idx % keep_every > 0:
            continue

        axes.PlotImageInZ(
            frame,
            all_rows,
            all_cols,
            n_levels,
            t_idx,
            scaling,
            min_intensity=min_intensity,
            intensity_range=intensity_range,
            alpha=_MILLEFEUILLE_ALPHA,
            cmap="gray",
        )

        if (cell_contours is not None) or with_cell_labels or (tracks is not None):
            if cell_contours is None:
                contours = None
            else:
                contours = cell_contours[t_idx]
            if cell_frames is None:
                cell_frame = None
            else:
                cell_frame = cell_frames[t_idx]

            _ = AnnotateCells(
                frame,
                contours,
                with_cell_labels,
                cell_frame,
                tracks,
                axes,
                CellAnnotationStyle,
                elevation=scaling * (t_idx + 0.2),
            )

    axes.SetTimeAxisProperties(n_frames - 1)


def ShowFramesAsTunnels(
    frames: Sequence[array_t],
    axes: axes_t,
    /,
    *,
    keep_every: int = 2,
    iso_value: float = 0.5,
) -> None:
    """"""
    volume = nmpy.array(frames)
    n_frames_all = frames.__len__()

    ShowFirstFrameAsFloor(frames[0], axes)

    axes.PlotIsosurface(
        volume,
        iso_value,
        should_be_capped=True,
        keep_every=keep_every,
        cmap="rainbow",
        lw=1,
    )
    axes.SetTimeAxisProperties(n_frames_all - 1)


def ShowFirstFrameAsFloor(frame: array_t, axes: axes_t, /) -> None:
    """"""
    shape = frame.shape
    all_rows, all_cols = nmpy.meshgrid(range(shape[0]), range(shape[1]), indexing="ij")

    min_intensity = nmpy.amin(frame)
    max_intensity = nmpy.amax(frame)
    intensity_range = max(1, max_intensity - min_intensity)

    axes.PlotImageInZ(
        frame,
        all_rows,
        all_cols,
        100,
        0.0,
        0.0,
        min_intensity=min_intensity,
        intensity_range=intensity_range,
        alpha=0.2,
        cmap="gray",
    )



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import numbers as nmbr
import tempfile as temp
from pathlib import Path as path_t
from typing import Dict, Optional, Sequence, Union

import xlsxwriter as xlsx

from cell_tracking_BC.in_out.text.logger import LOGGER
from cell_tracking_BC.standard.uid import AlphaColumnFromIndex
from cell_tracking_BC.type.sequence import sequence_t


event_response_h = Dict[int, Optional[Sequence[float]]]
division_times_h = Dict[int, Sequence[int]]
death_time_h = Dict[int, int]


DASH_TYPES = (
    "solid",
    "round_dot",
    "square_dot",
    "dash",
    "dash_dot",
    "long_dash",
    "long_dash_dot",
    "long_dash_dot_dot",
)
N_DASH_TYPES = DASH_TYPES.__len__()


def SaveCellFeatureToXLSX(
    path: Union[str, path_t],
    sequence: sequence_t,
    /,
    *,
    feature: Union[str, Sequence[str]] = None,
) -> None:
    """"""
    if isinstance(path, str):
        path = path_t(path)
    if feature is None:
        features = sequence.available_cell_features
    elif isinstance(feature, str):
        features = (feature,)
    else:
        features = feature

    if path.exists():
        print(f"{path}: File (or folder) already exists...")
        path = path_t(temp.mkdtemp()) / path.name
        print(f"Using {path} instead")

    workbook = xlsx.Workbook(str(path))

    for feature in features:
        evolutions = sequence.FeatureEvolutionsAlongAllTracks(feature)
        first_label = tuple(evolutions.keys())[0]
        if not isinstance(evolutions[first_label][1][0], nmbr.Number):
            continue

        worksheet = workbook.add_worksheet(feature)
        rows_limits = {}
        for label, (track, evolution) in evolutions.items():
            worksheet.write_row(label - 1, track.root_time_point, evolution)
            rows_limits[label] = (
                track.root_time_point,
                track.root_time_point + evolution.__len__() - 1,
            )

        chart = workbook.add_chart({"type": "line"})
        for l_idx, (row, (min_col, max_col)) in enumerate(rows_limits.items()):
            min_col = AlphaColumnFromIndex(min_col)
            max_col = AlphaColumnFromIndex(max_col)
            chart.add_series(
                {
                    "name": str(row),
                    "values": f"='{feature}'!${min_col}${row}:${max_col}${row}",
                    "line": {
                        "width": 1.0,
                        "dash_type": DASH_TYPES[l_idx % N_DASH_TYPES],
                    },
                }
            )
        worksheet.insert_chart(f"A{max(rows_limits.keys()) + 1}", chart)

    workbook.close()


def SaveCellEventsToXLSX(
    path: Union[str, path_t],
    cell_division_frame_idc: Union[division_times_h, Dict[str, division_times_h]],
    cell_death_frame_idc: Union[death_time_h, Dict[str, death_time_h]],
    n_divisions: Union[int, Dict[str, int]],
    /,
    *,
    division_response: Union[
        event_response_h,
        Dict[str, event_response_h],
    ] = None,
    death_response: Union[
        event_response_h,
        Dict[str, event_response_h],
    ] = None,
) -> None:
    """"""
    if isinstance(path, str):
        path = path_t(path)
    if path.exists():
        print(f"{path}: File (or folder) already exists...")
        path = path_t(temp.mkdtemp()) / path.name
        print(f"Using {path} instead")

    workbook = xlsx.Workbook(str(path))

    for SaveEvents, event_time in zip(
        (_SaveDivisionEvents, _SaveDeathEvents),
        (cell_division_frame_idc, cell_death_frame_idc),
    ):
        if _IsSingleDictionary(event_time):
            SaveEvents(workbook, "", event_time)
        else:
            for suffix, contents in event_time.items():
                SaveEvents(workbook, "#" + suffix, contents)

    if isinstance(n_divisions, int) and _IsSingleDictionary(cell_death_frame_idc):
        _SaveEventCounts(workbook, "", n_divisions, cell_death_frame_idc)
    elif not (
        isinstance(n_divisions, int) or _IsSingleDictionary(cell_death_frame_idc)
    ):
        for suffix, dea_contents in cell_death_frame_idc.items():
            n_divisions_suffix = n_divisions[suffix]
            _SaveEventCounts(workbook, "#" + suffix, n_divisions_suffix, dea_contents)
    else:
        LOGGER.warn(
            f"{isinstance(n_divisions, int)}/{_IsSingleDictionary(cell_death_frame_idc)}: "
            f"Content type mismatch. Expected: Both single- (True), or both multi-content (False)."
        )

    for event, response in zip(
        ("division", "death"), (division_response, death_response)
    ):
        if response is None:
            continue

        if _IsSingleDictionary(response):
            _SaveEventResponse(event, workbook, "", response)
        else:
            for suffix, contents in response.items():
                _SaveEventResponse(event, workbook, "#" + suffix, contents)

    workbook.close()


def _SaveDivisionEvents(
    workbook: xlsx.Workbook,
    sheet_suffix: str,
    cell_division_frame_idc: division_times_h,
    /,
) -> None:
    """"""
    worksheet = workbook.add_worksheet(f"division times{sheet_suffix}")
    for label, divisions_idc in cell_division_frame_idc.items():
        if divisions_idc is not None:
            worksheet.write_row(label - 1, 0, divisions_idc)


def _SaveDeathEvents(
    workbook: xlsx.Workbook, sheet_suffix: str, cell_death_frame_idc: death_time_h, /
) -> None:
    """"""
    worksheet = workbook.add_worksheet(f"death time{sheet_suffix}")
    for label, death_idx in cell_death_frame_idc.items():
        if death_idx is not None:
            worksheet.write_number(label - 1, 0, death_idx)


def _SaveEventCounts(
    workbook: xlsx.Workbook,
    sheet_suffix: str,
    n_divisions: int,
    cell_death_frame_idc: death_time_h,
    /,
) -> None:
    """"""
    worksheet = workbook.add_worksheet(f"event counts{sheet_suffix}")

    n_deaths_pattern = 0
    n_deaths_track = 0
    for death_idx in cell_death_frame_idc.values():
        if death_idx is not None:
            if death_idx >= 0:
                n_deaths_pattern += 1
            else:
                n_deaths_track += 1

    for r_idx, (title, value) in enumerate(
        zip(
            ("divisions", "death (pattern)", "death (track)", "death"),
            (
                n_divisions,
                n_deaths_pattern,
                n_deaths_track,
                n_deaths_pattern + n_deaths_track,
            ),
        )
    ):
        worksheet.write_string(r_idx, 0, title)
        worksheet.write_number(r_idx, 1, value)


def _SaveEventResponse(
    event: str,
    workbook: xlsx.Workbook,
    sheet_suffix: str,
    responses: event_response_h,
    /,
) -> None:
    """"""
    name = f"{event} response{sheet_suffix}"
    worksheet = workbook.add_worksheet(name)
    rows_limits = {}
    for label, response in responses.items():
        if response is not None:
            worksheet.write_row(label - 1, 0, response)
            rows_limits[label] = response.__len__() - 1

    chart = workbook.add_chart({"type": "line"})
    for l_idx, (row, max_col) in enumerate(rows_limits.items()):
        max_col = AlphaColumnFromIndex(max_col)
        chart.add_series(
            {
                "name": str(row),
                "values": f"='{name}'!$A${row}:${max_col}${row}",
                "line": {"width": 1.0, "dash_type": DASH_TYPES[l_idx % N_DASH_TYPES]},
            }
        )
    worksheet.insert_chart(f"A{max(rows_limits.keys()) + 1}", chart)


def _IsSingleDictionary(dictionary: dict) -> bool:
    """"""
    return isinstance(tuple(dictionary.keys())[0], int)



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from typing import Sequence, Union

import matplotlib.pyplot as pypl
import numpy as nmpy


array_t = nmpy.ndarray


def MatShow(
    *arrays: Sequence[array_t],
    axes_title: Union[str, Sequence[str]] = None,
    figure_title: str = None,
    should_run_event_loop: bool = True,
    **kwargs,
) -> None:
    """"""
    if isinstance(arrays, array_t):
        arrays = (arrays,)
        n_arrays = 1
    else:
        n_arrays = arrays.__len__()
    if axes_title is None:
        axes_title = n_arrays * (None,)
    elif isinstance(axes_title, str):
        axes_title = (axes_title,)

    n_cols = int(nmpy.ceil(nmpy.sqrt(n_arrays)).item())
    n_rows = int(nmpy.ceil(n_arrays / n_cols).item())
    figure, all_axes = pypl.subplots(nrows=n_rows, ncols=n_cols, squeeze=False)
    all_axes = tuple(_axs for _sub in all_axes for _axs in _sub)
    for axes, array, title in zip(all_axes, arrays, axes_title):
        axes.matshow(array, **kwargs)
        if title is not None:
            axes.set_title(title)

    if (n_axes := n_rows * n_cols) > n_arrays:
        for axes in all_axes[(n_arrays - n_axes) :]:
            axes.set_visible(False)

    if figure_title is not None:
        figure.suptitle(figure_title)
    figure.tight_layout(pad=1.04)

    if should_run_event_loop:
        pypl.show()



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from typing import Optional

import numpy as nmpy

import cell_tracking_BC.in_out.graphics.generic.three_d as thrd
from cell_tracking_BC.in_out.file.archiver import archiver_t
from cell_tracking_BC.in_out.graphics.context import axes_3d_t, context_t, figure_t
from cell_tracking_BC.in_out.graphics.dbe.matplotlib.context import DRAWING_CONTEXT
from cell_tracking_BC.type.sequence import sequence_t
from cell_tracking_BC.type.track import unstructured_track_t


_LOW_AFFINITY = 0.75
_MATPLOTLIB_COLORS = ("b", "g", "r", "c", "m", "y", "k")


def ShowTracking2D(
    sequence: sequence_t,
    /,
    *,
    version: str = None,
    with_segmentation: bool = True,
    with_cell_labels: bool = True,
    with_track_labels: bool = True,
    with_ticks: bool = True,
    with_colorbar: bool = True,
    prepare_only: bool = False,
    interactively: bool = True,
    in_main_thread: bool = True,
    figure_name: str = "tracking 2D",
    dbe: context_t = DRAWING_CONTEXT,
    archiver: archiver_t = None,
) -> Optional[figure_t]:
    """"""
    output = None

    viewer = dbe.t_viewer_2d_t.NewForSequence(
        sequence,
        dbe,
        version=version,
        with_segmentation=with_segmentation,
        with_cell_labels=with_cell_labels,
        with_track_labels=with_track_labels,
        with_ticks=with_ticks,
        with_colorbar=with_colorbar,
        figure_name=figure_name,
        archiver=archiver,
    )

    if prepare_only:
        output = viewer
    else:
        viewer.Show(interactively=interactively, in_main_thread=in_main_thread)

    return output


def ShowTracking3D(
    sequence: sequence_t,
    /,
    *,
    with_track_labels: bool = True,
    with_cell_labels: bool = True,
    prepare_only: bool = False,
    interactively: bool = True,
    in_main_thread: bool = True,
    dbe: context_t = DRAWING_CONTEXT,
) -> Optional[figure_t]:
    """"""
    output = None

    figure, axes = dbe.figure_3d_t.NewFigureAndAxes()
    axes: axes_3d_t
    colormap = axes.AddColormapFromMilestones(
        "Tracking Affinity", ((0.0, "black"), (0.75, "red"), (1.0, "blue"))
    )

    thrd.ShowFirstFrameAsFloor(sequence.cell_frames[0], axes)

    time_scaling = axes.__class__.TimeScaling(sequence.shape, sequence.length)
    for t_idx, track in enumerate(sequence.tracks):
        low_affinities = tuple(_ffn < _LOW_AFFINITY for _ffn in track.affinities)
        low_fraction = nmpy.count_nonzero(low_affinities) / (
            0.3 * low_affinities.__len__()
        )
        color = colormap(1.0 - min(1.0, low_fraction))

        for piece in track.Pieces():
            rows, cols, times, *labels = piece.AsRowsColsTimes(
                with_labels=with_cell_labels
            )
            times = tuple(time_scaling * _tme for _tme in times)

            axes.PlotLines(rows, cols, times, color=color)

            if with_cell_labels:
                for row, col, time, label in zip(rows, cols, times, labels[0]):
                    axes.PlotText(
                        row,
                        col,
                        time,
                        str(label),
                        fontsize="x-small",
                        color=color,
                    )
            if with_track_labels and (piece.label is not None):
                axes.PlotText(
                    rows[-1],
                    cols[-1],
                    times[-1] + 0.25,
                    str(piece.label),
                    fontsize="x-small",
                    color=color,
                )

    axes.SetTimeAxisProperties(sequence.length - 1)

    if prepare_only:
        output = figure
    else:
        figure.Show(interactively=interactively, in_main_thread=in_main_thread)

    return output


def ShowUnstructuredTracking3D(
    sequence: sequence_t,
    /,
    *,
    with_cell_labels: bool = True,
    prepare_only: bool = False,
    interactively: bool = True,
    in_main_thread: bool = True,
    dbe: context_t = DRAWING_CONTEXT,
) -> Optional[figure_t]:
    """"""
    output = None

    invalids = sequence.tracks.invalids[unstructured_track_t]
    if (invalids is None) or (invalids.__len__() == 0):
        return

    figure, axes = dbe.figure_3d_t.NewFigureAndAxes()
    axes: axes_3d_t

    colors = _MATPLOTLIB_COLORS
    for t_idx, (track, _) in enumerate(invalids):
        color_idx = t_idx % colors.__len__()

        for time_point, *cells, _ in track.segments_iterator:
            rows, cols = tuple(zip(*(_cll.centroid.tolist() for _cll in cells)))
            times = (time_point, time_point + 1)

            axes.PlotLines(rows, cols, times, colors[color_idx])

            if with_cell_labels:
                labels = tuple(_cll.label for _cll in cells)
                for row, col, time, label in zip(rows, cols, times, labels[0]):
                    axes.PlotText(
                        row,
                        col,
                        time,
                        str(label),
                        fontsize="x-small",
                        color=colors[color_idx],
                    )

    axes.SetTimeAxisProperties(sequence.length - 1)

    if prepare_only:
        output = figure
    else:
        figure.Show(interactively=interactively, in_main_thread=in_main_thread)

    return output



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import dataclasses as dcls
from typing import Dict, Optional, Sequence, Tuple

import numpy as nmpy
import scipy.optimize as spop
from scipy.spatial import distance as dstc

from cell_tracking_BC.task.jaccard import PseudoJaccard
from cell_tracking_BC.task.tracking.main import tracker_t
from cell_tracking_BC.type.cell import cell_t


array_t = nmpy.ndarray


@dcls.dataclass(repr=False, eq=False)
class gmb_tracker_t(tracker_t):
    """
    gmb: Global-Matching-Based tracking
    """

    shape: Sequence[int] = None
    min_jaccard: float = 0.75
    division_min_jaccard: float = None
    labeled_map: array_t = dcls.field(init=False, default=None)
    next_labeled_map: array_t = dcls.field(init=False, default=None)
    tp1_to_t_associations: Dict[int, int] = dcls.field(init=False, default=None)

    def __post_init__(self) -> None:
        """"""
        if self.division_min_jaccard is None:
            self.division_min_jaccard = self.min_jaccard

    def DiscoverNextCells(self, next_cells: Sequence[cell_t], /) -> None:
        """"""
        super().DiscoverNextCells(next_cells)

        cells_map = nmpy.zeros(self.shape, dtype=nmpy.uint64)
        for c_idx, cell in enumerate(next_cells, start=1):
            cell_map = cell.Map(self.shape, as_boolean=True)
            cells_map[cell_map] = c_idx
        self.next_labeled_map = cells_map

        if self.labeled_map is not None:
            cells_idc = nmpy.fromiter(
                range(1, self.cells.__len__() + 1), dtype=nmpy.uint64
            )
            next_cells_idc = nmpy.fromiter(
                range(1, self.next_cells.__len__() + 1), dtype=nmpy.uint64
            )
            cells_idc.shape = (-1, 1)
            next_cells_idc.shape = (-1, 1)
            _PseudoJaccard = lambda idx_1, idx_2: PseudoJaccard(
                self.labeled_map, self.next_labeled_map, idx_1, idx_2
            )
            pairwise_jaccards = dstc.cdist(
                cells_idc, next_cells_idc, metric=_PseudoJaccard
            )

            associations = {}
            for min_jaccard in (self.min_jaccard, self.division_min_jaccard):
                row_ind, col_ind = spop.linear_sum_assignment(1.0 - pairwise_jaccards)
                valid_idc = pairwise_jaccards[row_ind, col_ind] > min_jaccard
                if not nmpy.any(valid_idc):
                    break
                affinities = pairwise_jaccards[row_ind[valid_idc], col_ind[valid_idc]]
                values = zip(row_ind[valid_idc], affinities)
                new_associations = dict(zip(col_ind[valid_idc], values))
                associations.update(new_associations)
                pairwise_jaccards[:, col_ind[valid_idc]] = 0.0

            self.tp1_to_t_associations = associations

    def PredecessorOfCell(
        self, next_cell: cell_t, next_cell_idx: int, /
    ) -> Optional[Tuple[cell_t, float]]:
        """"""
        associated = self.tp1_to_t_associations.get(next_cell_idx)
        if associated is None:
            return None

        cell_idx, affinity = associated

        return self.cells[cell_idx], affinity

    def Advance(self) -> None:
        """"""
        super().Advance()
        self.labeled_map = self.next_labeled_map



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from typing import Sequence, Union

import numpy as nmpy
import plotext as pltx


array_t = nmpy.ndarray


def PlotHistogram(
    data: Union[array_t, Sequence[float]],
    /,
    *,
    n_bins: int = None,
    n_displayed_bins: int = 10,
    color: str = "blue",
    background_color: str = "black",
    axes_color: str = "green",
    title: str = None,
    xlabel: str = None,
    ylabel: str = None,
    width: int = 120,
    height: int = 20,
    should_use_numpy: bool = True,
    center_decimals: int = None,
) -> None:
    """"""
    if n_bins is None:
        n_bins = int(round(nmpy.sqrt(data.__len__())))

    pltx.clear_figure()

    if should_use_numpy:
        counts, edges = nmpy.histogram(data, bins=n_bins)

        centers = 0.5 * (edges[:-1] + edges[1:])
        if center_decimals is not None:
            centers = nmpy.around(centers, decimals=center_decimals)
            if center_decimals == 0:
                centers = centers.astype(nmpy.uint64)
        centers = tuple(str(_ctr) for _ctr in centers)

        pltx.bar(centers, counts.tolist(), color=color)
    else:
        n_displayed_bins = min(n_displayed_bins, n_bins)

        pltx.hist(data, n_bins, color=color)
        pltx.xfrequency(n_displayed_bins)

    pltx.canvas_color(background_color)
    pltx.axes_color(background_color)
    pltx.ticks_color(axes_color)

    if xlabel is not None:
        pltx.xlabel(xlabel)
    if ylabel is not None:
        pltx.ylabel(ylabel)
    if title is not None:
        pltx.title(title)

    terminal_width, _ = pltx.terminal_size()
    pltx.plot_size(min(width, terminal_width), height)

    pltx.show()



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from pathlib import Path as path_t
from typing import Callable, Union

import imageio as mgio
import mrc as mrci
import numpy as nmpy
import tifffile as tiff

import cell_tracking_BC.in_out.file.frame as frio
from cell_tracking_BC.in_out.graphics.context import context_t
from cell_tracking_BC.in_out.graphics.dbe.matplotlib.context import DRAWING_CONTEXT
from cell_tracking_BC.type.sequence import BoundingBoxSlices, sequence_t


array_t = nmpy.ndarray


# TODO: check the output format of the different functions. In particular, do they all output the time dimension as the
#       last one?
# TODO: Probably add (generic) parameters to specify eventual required hints for reading function such as number of
#       channels...


def SequenceByITK(path: path_t, /) -> array_t:
    """
    Shape: time*channel x row x col
    """
    return frio.FrameByITK(path)


def SequenceByIMAGEIO(path: path_t, /) -> array_t:
    """"""
    return mgio.volread(path)


def SequenceFromPath(
    path: path_t,
    /,
    *,
    SequenceLoading: Callable[[path_t], array_t] = SequenceByITK,
) -> array_t:
    """"""
    # TODO: make this function work in "every" cases (add a parameter about expected dimension arrangements), or
    #       remove the sequence loading functionality from cell-tracking-bc, leaving this task to the end user.
    if not (path.exists() and path.is_file()):
        raise ValueError(f"{path}: Not a path to an existing file")

    if (img_format := path.suffix[1:].lower()) in ("tif", "tiff"):
        output = tiff.imread(path)
    elif img_format in ("dv", "mrc"):
        # numpy.array: Because the returned value seems to be a read-only memory map
        output = nmpy.array(mrci.imread(str(path)))
        if output.ndim == 5:
            # Probably: time x channel x Z x Y x X while sequences are time x channel x (Z=1 x) Y x X, so one gets:
            # time x channel=1 x Z=actual channels x Y x X
            output = output[:, 0, :, :]
    else:
        output = SequenceLoading(path)

    return output


def SaveAnnotatedSequence(
    sequence: sequence_t,
    path: Union[str, path_t],
    /,
    *,
    channel: str = None,
    with_segmentation: bool = True,
    with_cell_labels: bool = True,
    with_track_labels: bool = True,
    as_sequence: bool = True,
    dbe: context_t = DRAWING_CONTEXT,
) -> None:
    """
    channel: None=segmentation channel

    See: cell_tracking_BC.in_out.graphics.matplotlib.generic_2d.s_viewer_2d_t.AsAnnotatedVolume
    """
    if isinstance(path, str):
        path = path_t(path)
    if channel is None:
        channel = sequence.cell_channel

    folder = path.parent
    basename = path.stem
    extension = path.suffix

    figure, axes = dbe.figure_2d_t.NewFigureAndAxes()
    viewer = dbe.s_viewer_2d_t.NewForChannels(
        sequence,
        dbe,
        channel=channel,
        with_segmentation=with_segmentation,
        with_cell_labels=with_cell_labels,
        with_track_labels=with_track_labels,
        with_ticks=False,
        with_colorbar=False,
        in_axes=axes,
    )

    output = None  # Cannot be initialized since content (not frame) shape is unknown
    for time_point in range(sequence.length):
        viewer.ShowFrame(time_point=time_point, should_draw=False)
        # draw_idle is not appropriate here
        figure.Update(gently=False)
        if as_sequence:
            content = figure.Content()
            if output is None:
                output = nmpy.empty((*content.shape, sequence.length), dtype=nmpy.uint8)
            output[..., time_point] = content
        else:
            figure.Save(folder / f"{basename}_{time_point}{extension}")

    figure.Close()  # To prevent remaining caught in event loop

    if as_sequence:
        row_slice, col_slice = BoundingBoxSlices(output)
        output = output[row_slice, col_slice, :, :]
        output = nmpy.moveaxis(output, (0, 1, 2, 3), (2, 3, 1, 0))
        output = output[:, nmpy.newaxis, :, :, :]
        tiff.imwrite(
            str(path),
            output,
            photometric="rgb",
            compression="deflate",
            planarconfig="separate",
            metadata={"axes": "XYZCT"},
        )



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from pathlib import Path

from setuptools import setup


AUTHOR = "Eric Debreuve"
E_MAIL = "eric.debreuve@univ-cotedazur.fr"

PYPI_NAME = "cell-tracking-bc"
DESCRIPTION = "Base Classes for Cell Tracking in Microscopy"
VERSION = (2022, 5)  # /!\ str(2021.10) -> "2021.1"
PY_VERSION = "3.8"

REPOSITORY_NAME = "cell-tracking-bc"
REPOSITORY_USER = "edebreuv"
REPOSITORY_SITE = "gitlab.inria.fr"
DOCUMENTATION_SITE = f"{REPOSITORY_USER}.gitlabpages.inria.fr"

IMPORT_NAME = "cell_tracking_BC"
PACKAGES = [
    IMPORT_NAME,
    f"{IMPORT_NAME}.in_out",
    f"{IMPORT_NAME}.in_out.file",
    f"{IMPORT_NAME}.in_out.graphics",
    f"{IMPORT_NAME}.in_out.graphics.dbe",
    f"{IMPORT_NAME}.in_out.graphics.dbe.matplotlib",
    f"{IMPORT_NAME}.in_out.graphics.dbe.mayavi",
    f"{IMPORT_NAME}.in_out.graphics.dbe.vedo",
    f"{IMPORT_NAME}.in_out.graphics.generic",
    f"{IMPORT_NAME}.in_out.text",
    f"{IMPORT_NAME}.standard",
    f"{IMPORT_NAME}.task",
    f"{IMPORT_NAME}.task.tracking",
    f"{IMPORT_NAME}.test",
    f"{IMPORT_NAME}.type",
]

HERE = Path(__file__).parent.resolve()


long_description = (HERE / "README.rst").read_text(encoding="utf-8")
repository_url = f"https://{REPOSITORY_SITE}/{REPOSITORY_USER}/{REPOSITORY_NAME}"
documentation_url = repository_url
version = f"{VERSION[0]}.{VERSION[1]}"

folders = [IMPORT_NAME]
for node in (HERE / IMPORT_NAME).rglob("*"):
    if node.is_dir():
        node = node.relative_to(HERE)
        node = str(node).replace("/", ".")
        folders.append(node)
folders = sorted(folders)
if PACKAGES != folders:
    raise ValueError(
        f"Missing packages in setup:\n"
        f"    - Declared={PACKAGES}\n"
        f"    - Actual={folders}"
    )


if __name__ == "__main__":
    #
    setup(
        # --- OLD BUILD PROCEDURE
        # Remove: {IMPORT_NAME.lower()}.egg-info/ build/ dist/
        # python setup.py bdist_wheel
        # --- NEW BUILD PROCEDURE
        # Remove: {IMPORT_NAME.lower()}.egg-info/ dist/
        # python -m build
        # --- UPLOAD
        # twine upload dist/*
        # https://pypi.org/project/{PYPI_NAME}
        author=AUTHOR,
        author_email=E_MAIL,
        #
        name=PYPI_NAME,
        description=DESCRIPTION,
        long_description=long_description,
        long_description_content_type="text/x-rst",
        version=version,
        license="CeCILL-2.1",
        #
        classifiers=[
            "Topic :: Scientific/Engineering :: Image Recognition",
            "Intended Audience :: Science/Research",
            "License :: OSI Approved :: CEA CNRS Inria Logiciel Libre License, version 2.1 (CeCILL-2.1)",
            f"Programming Language :: Python :: {PY_VERSION}",
            "Development Status :: 4 - Beta",
        ],
        keywords="cell, tracking, microscopy, classes",
        #
        url=repository_url,
        project_urls={
            "Documentation": documentation_url,
            "Source": repository_url,
        },
        #
        packages=PACKAGES,
        python_requires=f">={PY_VERSION}",
        install_requires=[
            "imageio",
            "itk",
            "matplotlib",
            # "mayavi",  # Temporarily removed since mayavi does not compile
            "mrc",
            "networkx",
            "numpy",
            "pca-b-stream",
            "Pillow",
            "plotext",
            "rich",
            "scikit-image",
            "scipy",
            "tensorflow",
            "tifffile",
            "vedo",
            "xlsxwriter",
        ],
    )



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from typing import Callable, Optional, Sequence, Tuple, Union

import numpy as nmpy
import skimage.measure as msre

import cell_tracking_BC.standard.uid as uqid
from cell_tracking_BC.in_out.graphics.context import (
    _sg,
    annotation_h,
    axes_2d_t,
    axes_3d_t,
    rgb_color_h,
    rgba_color_h,
)
from cell_tracking_BC.type.cell import cell_t
from cell_tracking_BC.type.frame import frame_t
from cell_tracking_BC.type.segmentation import compartment_t
from cell_tracking_BC.type.sequence import sequence_h, sequence_t
from cell_tracking_BC.type.tracks import tracks_t


array_t = nmpy.ndarray


def AnnotateCells(
    frame: Union[array_t, frame_t],
    cell_contours: Optional[Sequence[array_t]],
    with_cell_labels: bool,
    cell_frame: Optional[frame_t],
    tracks: Optional[tracks_t],
    axes: Union[axes_2d_t, axes_3d_t],
    AnnotationStyle: _sg.cell_annotation_style_h,
    /,
    *,
    highlighted: int = -1,
    elevation: float = None,
    with_alpha_cell_uids: bool = True,
) -> Sequence[Tuple[int, Union[annotation_h, Sequence[annotation_h]]]]:
    """"""
    output = []

    if with_cell_labels or (tracks is not None):
        if elevation is None:
            axes.RemoveLinesAndAnnotations()

        if cell_frame is None:
            labeled = msre.label(frame, connectivity=1)
            cells = msre.regionprops(labeled)
        else:
            cells = cell_frame.cells
        assert hasattr(cells[0], "centroid") and hasattr(
            cells[0], "label"
        ), "Please contact developer about API change"

        for cell in cells:
            if elevation is None:
                position = nmpy.flipud(cell.centroid)
            else:
                position = (*cell.centroid, elevation)

            text = []
            if with_cell_labels:
                text.append(uqid.AlphaColumnFromLabel(cell.label))
            else:
                text.append("")
            if (tracks is None) or not isinstance(cell, cell_t):
                text.append("")
            else:
                labels = tracks.TrackLabelsContainingCell(cell, tolerant_mode=True)
                if labels is None:
                    text.append("?")
                else:
                    if labels.__len__() > 1:
                        labels = "\n".join(str(_lbl) for _lbl in labels)
                    else:
                        labels = str(labels[0])
                    text.append(labels)

            if with_alpha_cell_uids:
                text = "".join(text)
                additionals = AnnotationStyle(cell.label == highlighted, "\n" in text)
            else:
                if text[0].__len__() > 0:
                    multi_text = [(text[0], {})]
                else:
                    multi_text = []
                if text[1].__len__() > 0:
                    multi_text.extend(
                        (_pce, {"rotation": -90.0}) for _pce in text[1].split("\n")
                    )
                text = multi_text
                additionals = AnnotationStyle(cell.label == highlighted, False)

            annotation = axes.PlotCellAnnotation(position, text, **additionals)

            output.append((cell.label, annotation))

    # Leave this block after cell annotation since, if placed before, the (new) contours are considered as previous
    # artists and removed.
    if cell_contours is not None:
        if elevation is None:
            for contour in cell_contours:
                axes.PlotLines(
                    contour[:, 1],
                    contour[:, 0],
                    linestyle=":",
                    color=(0.0, 1.0, 1.0, 0.3),
                )
        else:
            for contour in cell_contours:
                heights = contour.shape[0] * [elevation]
                axes.PlotLines(
                    contour[:, 0],
                    contour[:, 1],
                    heights,
                    linestyle=":",
                    color=(0.0, 1.0, 1.0, 0.3),
                )

    return output


def CellContours(
    sequence: sequence_h, with_segmentation: bool, /
) -> Optional[Sequence[Sequence[array_t]]]:
    """"""
    if (
        with_segmentation
        and isinstance(sequence, sequence_t)
        and (sequence.segmentations is not None)
    ):
        output = []
        for segmentation in sequence.segmentations.CompartmentsWithVersion(
            compartment_t.CELL
        ):
            output.append(msre.find_contours(segmentation))
    else:
        output = None

    return output


def CellTracks(sequence: sequence_h, with_track_labels: bool, /) -> Optional[tracks_t]:
    """"""
    with_track_labels = (
        with_track_labels
        and isinstance(sequence, sequence_t)
        and (sequence.tracks is not None)
    )
    if with_track_labels:
        output = sequence.tracks
    else:
        output = None

    return output


def ColorAndAlpha(
    color: Union[str, rgb_color_h, rgba_color_h],
    NameToRGB: Callable[[str], rgb_color_h],
    /,
    *,
    convert_to_rgb: bool = False,
) -> Tuple[Union[str, rgb_color_h], Optional[float]]:
    """"""
    if (is_str := isinstance(color, str)) or (color.__len__() == 3):
        alpha = None
        if is_str and convert_to_rgb:
            color = NameToRGB(color)
    else:
        alpha = color[-1]
        color = color[:-1]

    return color, alpha


def ZeroOneValueToRGBAWithMilestones(
    value: Union[float, Sequence[float]],
    milestones: Sequence[Tuple[float, str]],
    NameToRGB: Callable[[str], rgb_color_h],
    /,
) -> Union[rgba_color_h, Sequence[rgba_color_h]]:
    """"""
    if isinstance(value, Sequence):
        return tuple(
            ZeroOneValueToRGBAWithMilestones(_vle, milestones, NameToRGB)
            for _vle in value
        )

    n_milestones = milestones.__len__()
    m_idx = 0
    while (m_idx < n_milestones) and (value > milestones[m_idx][0]):
        m_idx += 1
    if m_idx >= n_milestones:
        color = milestones[-1][1]
    elif value < milestones[m_idx][0]:
        if m_idx > 0:
            previous = m_idx - 1
        else:
            previous = 0
        interval = (milestones[previous][1], milestones[m_idx][1])
        interval = (NameToRGB(_clr) for _clr in interval)
        ratio = (value - milestones[previous][0]) / (
            milestones[m_idx][0] - milestones[previous][0]
        )
        color = tuple(
            ratio * _end + (1.0 - ratio) * _stt for _stt, _end in zip(*interval)
        )
    else:
        color = milestones[m_idx][1]

    if isinstance(color, str):
        color = NameToRGB(color)
    color = (*color, 1.0)

    return color
